package ru.dbotthepony.mc.otm.matter;

import com.mojang.blaze3d.platform.InputConstants;
import net.minecraft.ChatFormatting;
import net.minecraft.client.Minecraft;
import net.minecraft.server.dedicated.DedicatedServer;
import net.minecraft.server.level.ServerPlayer;
import net.minecraft.world.item.Item;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.item.Items;
import net.minecraft.world.item.crafting.Ingredient;
import net.minecraft.world.item.crafting.Recipe;
import net.minecraft.world.item.crafting.RecipeType;
import net.minecraftforge.event.entity.player.ItemTooltipEvent;
import net.minecraftforge.event.entity.player.PlayerEvent;
import net.minecraftforge.event.server.ServerStartedEvent;
import net.minecraftforge.eventbus.api.SubscribeEvent;
import net.minecraftforge.network.PacketDistributor;
import ru.dbotthepony.mc.otm.OverdriveThatMatters;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;
import ru.dbotthepony.mc.otm.capability.drive.IMatteryDrive;
import ru.dbotthepony.mc.otm.core.Fraction;
import ru.dbotthepony.mc.otm.menu.FormattingHelper;
import ru.dbotthepony.mc.otm.network.MatterRegistryPacket;
import ru.dbotthepony.mc.otm.network.MatteryNetworking;
import ru.dbotthepony.mc.otm.storage.ItemStackWrapper;

import javax.annotation.Nullable;
import java.math.MathContext;
import java.util.*;

public class MatterRegistry {
	public static final Map<Item, Fraction> INITIAL_ITEMS = new HashMap<>();
	public static final Map<Item, Fraction> ITEMS = new HashMap<>();

	public static Fraction getMatterValue(Item item) {
		return ITEMS.getOrDefault(item, INITIAL_ITEMS.getOrDefault(item, Fraction.ZERO));
	}

	public static boolean hasMatterValue(Item item) {
		return ITEMS.containsKey(item) || INITIAL_ITEMS.containsKey(item);
	}

	public static boolean hasMatterValue(ItemStack item) {
		if (item.getCapability(MatteryCapability.DRIVE).isPresent() || item.getCapability(MatteryCapability.MATTER).isPresent())
			return true;

		return hasMatterValue(item.getItem());
	}

	public static boolean canDecompose(ItemStack stack) {
		if (stack.getCapability(MatteryCapability.DRIVE).isPresent() && stack.getCapability(MatteryCapability.DRIVE).resolve().get().getStoredCount().compareTo(Fraction.ZERO) > 0)
			return false;

		if (stack.getCapability(MatteryCapability.MATTER).isPresent() && stack.getCapability(MatteryCapability.MATTER).resolve().get().getStoredMatter().compareTo(Fraction.ZERO) > 0)
			return false;

		return hasMatterValue(stack.getItem());
	}

	private static Fraction getMatterValue(ItemStack stack, int level) {
		if (level >= 100) {
			return Fraction.ZERO;
		}

		Item item = stack.getItem();
		var matter = getMatterValue(item);

		if (item.isDamageable(stack)) {
			matter = matter.times(Fraction.ONE.minus(new Fraction(item.getDamage(stack)).div(new Fraction(item.getMaxDamage(stack)))));
		}

		var cap1 = stack.getCapability(MatteryCapability.DRIVE).resolve();
		var cap2 = stack.getCapability(MatteryCapability.MATTER).resolve();

		if (cap1.isPresent() && cap1.get().storageIdentity() == ItemStackWrapper.class) {
			for (var stored : ((IMatteryDrive<ItemStackWrapper>) cap1.get()).getStacks()) {
				matter = matter.plus(getMatterValue(stored.stack().stack(), level + 1));
			}
		}

		if (cap2.isPresent()) {
			matter = matter.plus(cap2.get().getStoredMatter());
		}

		return matter;
	}

	public static Fraction getMatterValue(ItemStack stack) {
		return getMatterValue(stack, 0);
	}

	public static void registerInitialItems() {
		// basic things
		INITIAL_ITEMS.put(Items.DIRT, new Fraction("0.0013"));
		INITIAL_ITEMS.put(Items.DIRT_PATH, new Fraction("0.0013"));
		INITIAL_ITEMS.put(Items.GRASS_BLOCK, new Fraction("0.0016"));

		INITIAL_ITEMS.put(Items.POINTED_DRIPSTONE, new Fraction("0.0007"));

		INITIAL_ITEMS.put(Items.STONE, new Fraction("0.001"));
		INITIAL_ITEMS.put(Items.COBBLESTONE, new Fraction("0.001"));
		INITIAL_ITEMS.put(Items.MOSSY_COBBLESTONE, new Fraction("0.001"));
		INITIAL_ITEMS.put(Items.NETHERRACK, new Fraction("0.00027"));
		INITIAL_ITEMS.put(Items.GRANITE, new Fraction("0.0011"));
		INITIAL_ITEMS.put(Items.DIORITE, new Fraction("0.0013"));
		INITIAL_ITEMS.put(Items.ANDESITE, new Fraction("0.0012"));
		INITIAL_ITEMS.put(Items.BASALT, new Fraction("0.0013"));

		INITIAL_ITEMS.put(Items.SOUL_SAND, new Fraction("0.00087"));
		INITIAL_ITEMS.put(Items.SOUL_SOIL, new Fraction("0.00087"));

		INITIAL_ITEMS.put(Items.ICE, new Fraction("0.0002"));

		INITIAL_ITEMS.put(Items.GRAVEL, new Fraction("0.001"));
		INITIAL_ITEMS.put(Items.FLINT, new Fraction("0.001"));

		INITIAL_ITEMS.put(Items.SAND, new Fraction("0.0005"));
		INITIAL_ITEMS.put(Items.RED_SAND, new Fraction("0.0005"));

		INITIAL_ITEMS.put(Items.TUFF, new Fraction("0.0007"));
		INITIAL_ITEMS.put(Items.DEEPSLATE, new Fraction("0.0014"));
		INITIAL_ITEMS.put(Items.COBBLED_DEEPSLATE, new Fraction("0.0014"));

		INITIAL_ITEMS.put(Items.BLACKSTONE, new Fraction("0.0015"));

		INITIAL_ITEMS.put(Items.INFESTED_STONE, new Fraction("0.063"));
		INITIAL_ITEMS.put(Items.INFESTED_COBBLESTONE, new Fraction("0.063"));
		INITIAL_ITEMS.put(Items.INFESTED_STONE_BRICKS, new Fraction("0.063"));
		INITIAL_ITEMS.put(Items.INFESTED_MOSSY_STONE_BRICKS, new Fraction("0.063"));
		INITIAL_ITEMS.put(Items.INFESTED_CRACKED_STONE_BRICKS, new Fraction("0.063"));
		INITIAL_ITEMS.put(Items.INFESTED_CHISELED_STONE_BRICKS, new Fraction("0.063"));
		INITIAL_ITEMS.put(Items.INFESTED_DEEPSLATE, new Fraction("0.067"));

		// affected by luck enchantment, thus can be duped
		// INITIAL_ITEMS.put(Items.GILDED_BLACKSTONE, new Fraction("0.037"));
		INITIAL_ITEMS.put(Items.GILDED_BLACKSTONE, new Fraction("0.34"));

		// assuming it is very light since it is floating
		INITIAL_ITEMS.put(Items.END_STONE, new Fraction("0.0004"));

		INITIAL_ITEMS.put(Items.OBSIDIAN, new Fraction("0.038"));

		// metallic / chemical things
		INITIAL_ITEMS.put(Items.COAL, new Fraction("0.005"));
		INITIAL_ITEMS.put(Items.LAPIS_LAZULI, new Fraction("0.0042"));
		INITIAL_ITEMS.put(Items.REDSTONE, new Fraction("0.004"));
		INITIAL_ITEMS.put(Items.CHARCOAL, new Fraction("0.005"));
		INITIAL_ITEMS.put(Items.DIAMOND, new Fraction("0.5"));
		INITIAL_ITEMS.put(Items.EMERALD, new Fraction("0.46"));
		INITIAL_ITEMS.put(Items.IRON_INGOT, new Fraction("0.03"));
		INITIAL_ITEMS.put(Items.COPPER_INGOT, new Fraction("0.014"));
		INITIAL_ITEMS.put(Items.GOLD_INGOT, new Fraction("0.32"));
		INITIAL_ITEMS.put(Items.NETHERITE_SCRAP, new Fraction("1.2"));
		INITIAL_ITEMS.put(Items.QUARTZ, new Fraction("0.008"));
		INITIAL_ITEMS.put(Items.GLOWSTONE_DUST, new Fraction("0.007"));
		INITIAL_ITEMS.put(Items.AMETHYST_SHARD, new Fraction("0.034"));

		// living things
		INITIAL_ITEMS.put(Items.MOSS_BLOCK, new Fraction("0.0012"));

		INITIAL_ITEMS.put(Items.OAK_LOG, new Fraction("0.005"));
		INITIAL_ITEMS.put(Items.SPRUCE_LOG, new Fraction("0.005"));
		INITIAL_ITEMS.put(Items.BIRCH_LOG, new Fraction("0.005"));
		INITIAL_ITEMS.put(Items.JUNGLE_LOG, new Fraction("0.005"));
		INITIAL_ITEMS.put(Items.ACACIA_LOG, new Fraction("0.005"));
		INITIAL_ITEMS.put(Items.DARK_OAK_LOG, new Fraction("0.005"));

		INITIAL_ITEMS.put(Items.BAMBOO, new Fraction("0.00085"));

		INITIAL_ITEMS.put(Items.CRIMSON_PLANKS, new Fraction("0.003"));
		INITIAL_ITEMS.put(Items.WARPED_PLANKS, new Fraction("0.003"));

		INITIAL_ITEMS.put(Items.STRIPPED_OAK_LOG, new Fraction("0.005"));
		INITIAL_ITEMS.put(Items.STRIPPED_SPRUCE_LOG, new Fraction("0.005"));
		INITIAL_ITEMS.put(Items.STRIPPED_BIRCH_LOG, new Fraction("0.005"));
		INITIAL_ITEMS.put(Items.STRIPPED_JUNGLE_LOG, new Fraction("0.005"));
		INITIAL_ITEMS.put(Items.STRIPPED_ACACIA_LOG, new Fraction("0.005"));
		INITIAL_ITEMS.put(Items.STRIPPED_DARK_OAK_LOG, new Fraction("0.005"));

		INITIAL_ITEMS.put(Items.CACTUS, new Fraction("0.0039"));
		INITIAL_ITEMS.put(Items.PUMPKIN, new Fraction("0.005"));
		INITIAL_ITEMS.put(Items.MELON, new Fraction("0.005"));

		// flowers!
		final Fraction flower_value = new Fraction("0.0031");
		INITIAL_ITEMS.put(Items.DANDELION, flower_value);
		INITIAL_ITEMS.put(Items.POPPY, flower_value);
		INITIAL_ITEMS.put(Items.BLUE_ORCHID, flower_value);
		INITIAL_ITEMS.put(Items.ALLIUM, flower_value);
		INITIAL_ITEMS.put(Items.AZURE_BLUET, flower_value);
		INITIAL_ITEMS.put(Items.RED_TULIP, flower_value);
		INITIAL_ITEMS.put(Items.ORANGE_TULIP, flower_value);
		INITIAL_ITEMS.put(Items.WHITE_TULIP, flower_value);
		INITIAL_ITEMS.put(Items.PINK_TULIP, flower_value);
		INITIAL_ITEMS.put(Items.OXEYE_DAISY, flower_value);
		INITIAL_ITEMS.put(Items.CORNFLOWER, flower_value);
		INITIAL_ITEMS.put(Items.LILY_OF_THE_VALLEY, flower_value);

		final Fraction big_flower_value = new Fraction("0.0042");
		INITIAL_ITEMS.put(Items.SUNFLOWER, big_flower_value);
		INITIAL_ITEMS.put(Items.LILAC, big_flower_value);
		INITIAL_ITEMS.put(Items.ROSE_BUSH, big_flower_value);
		INITIAL_ITEMS.put(Items.PEONY, big_flower_value);

		INITIAL_ITEMS.put(Items.SPORE_BLOSSOM, new Fraction("0.0067"));
		INITIAL_ITEMS.put(Items.BROWN_MUSHROOM, new Fraction("0.0034"));
		INITIAL_ITEMS.put(Items.RED_MUSHROOM, new Fraction("0.0034"));
		INITIAL_ITEMS.put(Items.CRIMSON_FUNGUS, new Fraction("0.004"));
		INITIAL_ITEMS.put(Items.WARPED_FUNGUS, new Fraction("0.004"));

		// grass
		final Fraction grass_value = new Fraction("0.0024");
		INITIAL_ITEMS.put(Items.CRIMSON_ROOTS, grass_value);
		INITIAL_ITEMS.put(Items.WARPED_ROOTS, grass_value);
		INITIAL_ITEMS.put(Items.NETHER_SPROUTS, new Fraction("0.002"));
		INITIAL_ITEMS.put(Items.WEEPING_VINES, grass_value);
		INITIAL_ITEMS.put(Items.TWISTING_VINES, grass_value);
		INITIAL_ITEMS.put(Items.HANGING_ROOTS, grass_value);
		INITIAL_ITEMS.put(Items.TALL_GRASS, new Fraction("0.0035"));
		INITIAL_ITEMS.put(Items.LARGE_FERN, new Fraction("0.0035"));
		INITIAL_ITEMS.put(Items.GRASS, grass_value);
		INITIAL_ITEMS.put(Items.VINE, grass_value);
		INITIAL_ITEMS.put(Items.LILY_PAD, grass_value);
		INITIAL_ITEMS.put(Items.BIG_DRIPLEAF, grass_value);
		INITIAL_ITEMS.put(Items.SMALL_DRIPLEAF, grass_value);
		INITIAL_ITEMS.put(Items.SEA_PICKLE, grass_value);
		INITIAL_ITEMS.put(Items.SEAGRASS, grass_value);
		INITIAL_ITEMS.put(Items.FERN, grass_value);
		INITIAL_ITEMS.put(Items.DEAD_BUSH, new Fraction("0.0017"));
		INITIAL_ITEMS.put(Items.GLOW_LICHEN, new Fraction("0.0026"));
		INITIAL_ITEMS.put(Items.AZALEA, new Fraction("0.0018"));
		INITIAL_ITEMS.put(Items.FLOWERING_AZALEA, grass_value);

		// living plant (not very blocky) things
		// saplings
		INITIAL_ITEMS.put(Items.OAK_SAPLING, new Fraction("0.0035"));
		INITIAL_ITEMS.put(Items.SPRUCE_SAPLING, new Fraction("0.0045"));
		INITIAL_ITEMS.put(Items.BIRCH_SAPLING, new Fraction("0.0035"));
		INITIAL_ITEMS.put(Items.JUNGLE_SAPLING, new Fraction("0.0048"));
		INITIAL_ITEMS.put(Items.ACACIA_SAPLING, new Fraction("0.0033"));
		INITIAL_ITEMS.put(Items.DARK_OAK_SAPLING, new Fraction("0.005"));

		INITIAL_ITEMS.put(Items.WHEAT_SEEDS, new Fraction("0.0007"));
		INITIAL_ITEMS.put(Items.BEETROOT_SEEDS, new Fraction("0.0007"));
		// INITIAL_ITEMS.put(Items.MELON_SEEDS, new Fraction("0.0013"));
		INITIAL_ITEMS.put(Items.WHEAT, new Fraction("0.0016"));
		INITIAL_ITEMS.put(Items.NETHER_WART, new Fraction("0.0017"));
		INITIAL_ITEMS.put(Items.CARROT, new Fraction("0.0019"));
		INITIAL_ITEMS.put(Items.POTATO, new Fraction("0.0021"));
		INITIAL_ITEMS.put(Items.BEETROOT, new Fraction("0.0021"));
		INITIAL_ITEMS.put(Items.MELON_SLICE, new Fraction("0.0008"));

		INITIAL_ITEMS.put(Items.COCOA_BEANS, new Fraction("0.00035"));
		INITIAL_ITEMS.put(Items.HONEYCOMB, new Fraction("0.0014"));
		INITIAL_ITEMS.put(Items.SUGAR_CANE, new Fraction("0.004"));
		INITIAL_ITEMS.put(Items.APPLE, new Fraction("0.0061"));
		INITIAL_ITEMS.put(Items.SWEET_BERRIES, new Fraction("0.0035"));
		INITIAL_ITEMS.put(Items.GLOW_BERRIES, new Fraction("0.0041"));
		INITIAL_ITEMS.put(Items.KELP, new Fraction("0.0009"));

		// living animal things
		INITIAL_ITEMS.put(Items.STRING, new Fraction("0.0006"));
		INITIAL_ITEMS.put(Items.COBWEB, new Fraction("0.0006"));
		INITIAL_ITEMS.put(Items.INK_SAC, new Fraction("0.0009"));
		INITIAL_ITEMS.put(Items.SPIDER_EYE, new Fraction("0.001"));
		INITIAL_ITEMS.put(Items.FEATHER, new Fraction("0.0007"));
		INITIAL_ITEMS.put(Items.GUNPOWDER, new Fraction("0.003"));
		INITIAL_ITEMS.put(Items.LEATHER, new Fraction("0.0065"));
		INITIAL_ITEMS.put(Items.BONE, new Fraction("0.004"));
		INITIAL_ITEMS.put(Items.ENDER_PEARL, new Fraction("0.041"));
		INITIAL_ITEMS.put(Items.GHAST_TEAR, new Fraction("0.023"));
		INITIAL_ITEMS.put(Items.BLAZE_ROD, new Fraction("0.02"));
		INITIAL_ITEMS.put(Items.SLIME_BALL, new Fraction("0.0015"));
		INITIAL_ITEMS.put(Items.EGG, new Fraction("0.0011"));

		INITIAL_ITEMS.put(Items.PORKCHOP, new Fraction("0.0047"));
		INITIAL_ITEMS.put(Items.BEEF, new Fraction("0.0047"));
		INITIAL_ITEMS.put(Items.MUTTON, new Fraction("0.004"));

		INITIAL_ITEMS.put(Items.PUFFERFISH, new Fraction("0.013"));
		INITIAL_ITEMS.put(Items.COD, new Fraction("0.013"));
		INITIAL_ITEMS.put(Items.SALMON, new Fraction("0.013"));
		INITIAL_ITEMS.put(Items.TROPICAL_FISH, new Fraction("0.013"));

		// building items
		INITIAL_ITEMS.put(Items.CLAY_BALL, new Fraction("0.0006"));
		INITIAL_ITEMS.put(Items.SNOWBALL, new Fraction("0.00041"));

		// loot
		INITIAL_ITEMS.put(Items.TOTEM_OF_UNDYING, new Fraction("1.47"));
		INITIAL_ITEMS.put(Items.TRIDENT, new Fraction("1.35"));

		/*
		INITIAL_ITEMS.put(Items.WHITE_WOOL, new Fraction("0.004"));
		INITIAL_ITEMS.put(Items.ORANGE_WOOL, new Fraction("0.004"));
		INITIAL_ITEMS.put(Items.MAGENTA_WOOL, new Fraction("0.004"));
		INITIAL_ITEMS.put(Items.LIGHT_BLUE_WOOL, new Fraction("0.004"));
		INITIAL_ITEMS.put(Items.YELLOW_WOOL, new Fraction("0.004"));
		INITIAL_ITEMS.put(Items.LIME_WOOL, new Fraction("0.004"));
		INITIAL_ITEMS.put(Items.PINK_WOOL, new Fraction("0.004"));
		INITIAL_ITEMS.put(Items.GRAY_WOOL, new Fraction("0.004"));
		INITIAL_ITEMS.put(Items.LIGHT_GRAY_WOOL, new Fraction("0.004"));
		INITIAL_ITEMS.put(Items.CYAN_WOOL, new Fraction("0.004"));
		INITIAL_ITEMS.put(Items.PURPLE_WOOL, new Fraction("0.004"));
		INITIAL_ITEMS.put(Items.BLUE_WOOL, new Fraction("0.004"));
		INITIAL_ITEMS.put(Items.BROWN_WOOL, new Fraction("0.004"));
		INITIAL_ITEMS.put(Items.GREEN_WOOL, new Fraction("0.004"));
		INITIAL_ITEMS.put(Items.RED_WOOL, new Fraction("0.004"));
		INITIAL_ITEMS.put(Items.BLACK_WOOL, new Fraction("0.004"));
		 */

		// INITIAL_ITEMS.put(Items.CRIMSON_LOG, new Fraction("0.005"));
		// INITIAL_ITEMS.put(Items.WARPED_LOG, new Fraction("0.005"));
	}

	private static void registerPostItem(Item item, Item base, Fraction multiplier) {
		Fraction value = getMatterValue(base);

		if (!value.equals(Fraction.ZERO)) {
			ITEMS.put(item, value.times(multiplier));
		}
	}

	public static void registerPostItems() {
		registerPostItem(Items.CHIPPED_ANVIL, Items.ANVIL, new Fraction("0.8"));
		registerPostItem(Items.DAMAGED_ANVIL, Items.ANVIL, new Fraction("0.5"));
	}

	public static final MathContext ROUND_RULES = MatteryCapability.ROUND_RULES;

	private static final HashMap<Item, ArrayList<Recipe<?>>> usages = new HashMap<>();
	private static final HashMap<Item, ArrayList<Recipe<?>>> results = new HashMap<>();
	private static final HashMap<Item, Integer> seen_items = new HashMap<>();
	private static final HashMap<Item, Item> seen_items_child = new HashMap<>();

	private static boolean solve_occured = false;
	private static final Set<Item> defered_items = new HashSet<>();

	private static int stack_index = 0;
	private static final Item[] scan_stack = new Item[1000];

	@Nullable
	private static Fraction determineValue(Item item) {
		Fraction _get = ITEMS.get(item);

		if (_get != null) {
			return _get;
		}

		_get = INITIAL_ITEMS.get(item);

		if (_get != null) {
			return _get;
		}

		if (defered_items.contains(item)) {
			return null;
		}

		for (Item value : scan_stack) {
			if (value == null)
				break;

			if (value == item) {
				// defer, we hit an recursive recipe
				defered_items.add(item);
				return null;
			}
		}

		if (seen_items.containsKey(item)) {
			// OverdriveThatMatters.LOGGER.debug("Already seen: {}", item);
			seen_items.put(item, seen_items.get(item) + 1);

			if (seen_items_child.containsKey(item)) {
				for (Item child : seen_items_child.values()) {
					seen_items.put(child, seen_items.get(child) + 1);
				}
			}

			return Fraction.ZERO;
		}

		seen_items.put(item, 0);

		if (!results.containsKey(item)) {
			return Fraction.ZERO;
		}

		scan_stack[stack_index] = item;
		stack_index++;

		boolean defer_occured = false;

		Fraction smallest_possible_total = null;

		for (Recipe<?> recipe : results.get(item)) {
			ItemStack self = recipe.getResultItem();
			Fraction recipe_summ = Fraction.ZERO;

			boolean this_defered = false;

			for (Ingredient ingredient : recipe.getIngredients()) {
				ItemStack[] items = ingredient.getItems();

				Fraction smallest_possible = null;

				for (ItemStack stack : items) {
					if (!stack.isEmpty()) {
						Fraction determine = determineValue(stack.getItem());

						// if we hit an recursive recipe, defer it
						if (determine != null) {
							if (determine.compareTo(Fraction.ZERO) == 0) {
								// farewell
								OverdriveThatMatters.LOGGER.debug("Cannot determine matter value for {} because value of {} is unknown", item, stack.getItem());
								seen_items_child.put(item, stack.getItem());
								scan_stack[stack_index] = null;
								stack_index--;
								return Fraction.ZERO;
							}

							if (smallest_possible == null || smallest_possible.compareTo(determine) > 0)
								smallest_possible = determine;
						} else {
							OverdriveThatMatters.LOGGER.debug("Matter value for {} happened to be defered in calculations because {} is defered", item, stack.getItem());
							defer_occured = true;
							this_defered = true;
							smallest_possible = null;
							break;
						}
					}
				}

				if (this_defered)
					break;

				if (smallest_possible != null) {
					recipe_summ = recipe_summ.plus(smallest_possible);
				}
			}

			if (this_defered)
				continue;

			recipe_summ = recipe_summ.div(new Fraction(self.getCount()));

			if (smallest_possible_total == null || smallest_possible_total.compareTo(recipe_summ) > 0)
				smallest_possible_total = recipe_summ;
		}

		if (smallest_possible_total != null) {
			// ez
			ITEMS.put(item, smallest_possible_total);
			scan_stack[stack_index] = null;
			stack_index--;
			solve_occured = true;
			return smallest_possible_total;
		}

		if (defer_occured) {
			defered_items.add(item);
			scan_stack[stack_index] = null;
			stack_index--;
			OverdriveThatMatters.LOGGER.debug("Matter value for {} is defered completely until next iteration", item);
			return null;
		}

		scan_stack[stack_index] = null;
		stack_index--;
		return Fraction.ZERO;
	}

	private static void flood(List<? extends Recipe<?>> recipe_type) {
		for (Recipe<?> recipe : recipe_type) {
			ItemStack result = recipe.getResultItem();

			if (result.isEmpty())
				continue;

			ArrayList<Recipe<?>> results_get = results.computeIfAbsent(result.getItem(), k -> new ArrayList<>());

			if (results_get.contains(recipe))
				return;

			results_get.add(recipe);

			for (Ingredient ingredient :  recipe.getIngredients()) {
				ItemStack[] items = ingredient.getItems();

				for (ItemStack stack : items) {
					if (!stack.isEmpty()) {
						ArrayList<Recipe<?>> usages_get = usages.computeIfAbsent(stack.getItem(), k -> new ArrayList<>());

						if (!usages_get.contains(recipe)) {
							usages_get.add(recipe);
						}
					}
				}
			}
		}
	}

	private static boolean should_network = false;

	@SubscribeEvent
	public static void onClientLogin(PlayerEvent.PlayerLoggedInEvent event) {
		if (!should_network)
			return;

		MatteryNetworking.CHANNEL.send(PacketDistributor.PLAYER.with(() -> (ServerPlayer) event.getPlayer()), new MatterRegistryPacket(ITEMS));
	}

	@SubscribeEvent
	public static void onServerStarted(ServerStartedEvent event) {
		should_network = event.getServer() instanceof DedicatedServer;

		ITEMS.clear();
		usages.clear();
		results.clear();
		defered_items.clear();
		seen_items.clear();
		seen_items_child.clear();

		// scan for recipes and build linked tables
		flood(event.getServer().getRecipeManager().getAllRecipesFor(RecipeType.CRAFTING));
		flood(event.getServer().getRecipeManager().getAllRecipesFor(RecipeType.SMELTING));
		flood(event.getServer().getRecipeManager().getAllRecipesFor(RecipeType.STONECUTTING));
		flood(event.getServer().getRecipeManager().getAllRecipesFor(RecipeType.SMITHING));

		// expand the recipe tree and find references
		// just scan entire known tree
		solve_occured = true;

		for (int i = 0; i < 100 && solve_occured; i++)
			for (Item item : results.keySet()) {
				seen_items.clear();
				seen_items_child.clear();
				defered_items.clear();
				determineValue(item);
			}

		// clean up garbage
		usages.clear();
		results.clear();
		defered_items.clear();

		ArrayList<Map.Entry<Item, Integer>> required_resolve = new ArrayList<>(seen_items.entrySet());
		required_resolve.sort((a, b) -> Integer.compare(0, a.getValue().compareTo(b.getValue())));

		for (Map.Entry<Item, Integer> entry : required_resolve) {
			if (entry.getValue() > 0) {
				OverdriveThatMatters.LOGGER.debug("Resolving matter value of {} will unlock {} subcalls (reason: {})", entry.getKey(), entry.getValue(), seen_items_child.getOrDefault(entry.getKey(), Items.AIR));
			}
		}

		seen_items.clear();
		seen_items_child.clear();

		registerPostItems();
	}

	@SubscribeEvent
	public static void attachItemStackTextEvent(ItemTooltipEvent event) {
		if (InputConstants.isKeyDown(Minecraft.getInstance().getWindow().getWindow(), 340) || InputConstants.isKeyDown(Minecraft.getInstance().getWindow().getWindow(), 344)) {
			Fraction matter_value = getMatterValue(event.getItemStack());

			if (!matter_value.equals(Fraction.ZERO)) {
				event.getToolTip().add(FormattingHelper.formatMatterValue(matter_value).withStyle(ChatFormatting.AQUA));
			}
		}
	}
}
