package ru.dbotthepony.mc.otm.item

import net.minecraft.ChatFormatting
import net.minecraft.MethodsReturnNonnullByDefault
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.TranslatableComponent
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.Rarity
import net.minecraft.world.item.TooltipFlag
import net.minecraft.world.level.Level
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ICapabilityProvider
import net.minecraftforge.common.util.LazyOptional
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.matter.IMatterHandler
import ru.dbotthepony.mc.otm.capability.matter.IMatterHandler.MatterDirection
import ru.dbotthepony.mc.otm.core.Fraction
import ru.dbotthepony.mc.otm.core.Fraction.Companion.deserializeNBT
import ru.dbotthepony.mc.otm.menu.FormattingHelper
import javax.annotation.ParametersAreNonnullByDefault

@ParametersAreNonnullByDefault
@MethodsReturnNonnullByDefault
class ItemMatterCapacitor : Item {
	private inner class ItemMatterCapacitorCapability(private val stack: ItemStack) : ICapabilityProvider, IMatterHandler {
		private val resolver = LazyOptional.of<IMatterHandler> { this }

		override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
			return if (cap === MatteryCapability.MATTER) resolver.cast() else LazyOptional.empty()
		}

		private fun matter(): Fraction {
			val tag = stack.orCreateTag
			return if (tag.contains("otm_matter")) {
				deserializeNBT(tag["otm_matter"])
			} else Fraction.ZERO
		}

		private fun matter(value: Fraction) {
			stack.getOrCreateTag().put("otm_matter", value.serializeNBT())
		}

		override fun getStoredMatter(): Fraction {
			return if (isCreative) Fraction.LONG_MAX_VALUE else matter()
		}

		override fun getMaxStoredMatter(): Fraction {
			return storage
		}

		override fun getMissingMatter(): Fraction {
			return if (isCreative) Fraction.LONG_MAX_VALUE else super.getMissingMatter()
		}

		override fun receiveMatterOuter(howMuch: Fraction, simulate: Boolean): Fraction {
			return receiveMatterInner(howMuch, simulate)
		}

		override fun receiveMatterInner(howMuch: Fraction, simulate: Boolean): Fraction {
			if (isCreative) return howMuch
			val new = matter().plus(howMuch.min(maxInput)).min(storage)
			val diff = new.minus(matter())

			if (!simulate) {
				matter(new)
			}

			return diff
		}

		override fun extractMatterOuter(howMuch: Fraction, simulate: Boolean): Fraction {
			return extractMatterInner(howMuch, simulate)
		}

		override fun extractMatterInner(howMuch: Fraction, simulate: Boolean): Fraction {
			if (isCreative) return howMuch
			val new = matter().minus(howMuch.min(maxOutput)).moreThanZero()
			val diff = matter().minus(new)

			if (!simulate) {
				matter(new)
			}

			return diff
		}

		override fun getDirection(): MatterDirection {
			return MatterDirection.BIDIRECTIONAL
		}
	}

	val storage: Fraction
	private val isCreative: Boolean

	constructor(storage: Fraction) : super(Properties().stacksTo(1).tab(OverdriveThatMatters.CREATIVE_TAB)) {
		isCreative = false
		this.storage = storage
	}

	constructor() : super(Properties().stacksTo(1).tab(OverdriveThatMatters.CREATIVE_TAB).rarity(Rarity.EPIC)) {
		isCreative = true
		storage = MatteryCapability.LONG_MAX_VALUE
	}

	override fun appendHoverText(
		stack: ItemStack,
		p_41422_: Level?,
		p_41423_: MutableList<Component>,
		p_41424_: TooltipFlag
	) {
		super.appendHoverText(stack, p_41422_, p_41423_, p_41424_)
		if (isCreative) {
			p_41423_.add(INFINITE_STORAGE)
		} else {
			stack.getCapability(MatteryCapability.MATTER).ifPresent { cap: IMatterHandler ->
				p_41423_.add(
					TranslatableComponent(
						"otm.item.matter.normal",
						FormattingHelper.formatMatterValuePlain(cap.storedMatter),
						FormattingHelper.formatMatterValuePlain(storage)
					).withStyle(ChatFormatting.GRAY)
				)
			}
		}
	}

	override fun initCapabilities(stack: ItemStack, nbt: CompoundTag?): ICapabilityProvider {
		return ItemMatterCapacitorCapability(stack)
	}

	companion object {
		private val maxInput = MatteryCapability.LONG_MAX_VALUE
		private val maxOutput = MatteryCapability.LONG_MAX_VALUE

		private val INFINITE_STORAGE: Component =
			TranslatableComponent("otm.item.matter.infinite").withStyle(ChatFormatting.GRAY)
	}
}